// SPDX-License-Identifier: GPL-2.0-only
/*
 * Copyright Red Hat
 */

#include <linux/atomic.h>
#include <linux/blkdev.h>
#include <linux/mount.h>

#include "io-factory.h"
#include "logger.h"
#include "memory-alloc.h"

enum { BLK_FMODE = BLK_OPEN_READ | BLK_OPEN_WRITE };

/*
 * A kernel mode IO Factory object controls access to an index stored
 * on a block device.
 */
struct io_factory {
	struct block_device *bdev;
	struct bdev_handle *bdev_h;
	atomic_t ref_count;
};

void get_uds_io_factory(struct io_factory *factory)
{
	atomic_inc(&factory->ref_count);
}

static int get_block_device_from_name(const char *name,
				      struct bdev_handle **bdev_h)
{
	dev_t device;
	unsigned int major, minor;
	char dummy;
	const struct blk_holder_ops hops = { NULL };

	/* Extract the major/minor numbers */
	if (sscanf(name, "%u:%u%c", &major, &minor, &dummy) == 2) {
		device = MKDEV(major, minor);
		if (MAJOR(device) != major || MINOR(device) != minor) {
			*bdev_h = NULL;
			return uds_log_error_strerror(UDS_INVALID_ARGUMENT,
						      "%s is not a valid block device",
						      name);
		}
		*bdev_h = bdev_open_by_dev(device, BLK_FMODE, NULL, &hops);
	} else {
		*bdev_h = bdev_open_by_path(name, BLK_FMODE, NULL, &hops);
	}

	if (IS_ERR(*bdev_h)) {
		uds_log_error_strerror(-PTR_ERR(*bdev_h), "%s is not a block device", name);
		*bdev_h = NULL;
		return UDS_INVALID_ARGUMENT;
	}

	return UDS_SUCCESS;
}

int make_uds_io_factory(const char *path, struct io_factory **factory_ptr)
{
	int result;
	struct bdev_handle *bdev_h;
	struct io_factory *factory;

	result = get_block_device_from_name(path, &bdev_h);
	if (result != UDS_SUCCESS) {
		return result;
	}

	result = UDS_ALLOCATE(1, struct io_factory, __func__, &factory);
	if (result != UDS_SUCCESS) {
		bdev_release(bdev_h);
		return result;
	}

	factory->bdev_h = bdev_h;
	factory->bdev = bdev_h->bdev;
	atomic_set_release(&factory->ref_count, 1);

	*factory_ptr = factory;
	return UDS_SUCCESS;
}

int replace_uds_storage(struct io_factory *factory, const char *path)
{
	int result;
	struct bdev_handle *bdev_h;

	result = get_block_device_from_name(path, &bdev_h);
	if (result != UDS_SUCCESS) {
		return result;
	}

	factory->bdev = NULL;
	bdev_release(factory->bdev_h);
	factory->bdev_h = bdev_h;
	factory->bdev = bdev_h->bdev;
	return UDS_SUCCESS;
}

void put_uds_io_factory(struct io_factory *factory)
{
	if (atomic_add_return(-1, &factory->ref_count) <= 0) {
		factory->bdev=NULL;
		bdev_release(factory->bdev_h);
		UDS_FREE(factory);
	}
}

size_t get_uds_writable_size(struct io_factory *factory)
{
	return i_size_read(factory->bdev->bd_inode);
}

int make_uds_bufio(struct io_factory *factory,
		   off_t offset,
		   size_t block_size,
		   unsigned int reserved_buffers,
		   struct dm_bufio_client **client_ptr)
{
	struct dm_bufio_client *client;

	if (offset % SECTOR_SIZE != 0) {
		return uds_log_error_strerror(UDS_INCORRECT_ALIGNMENT,
					      "offset %zd not multiple of %d",
					      offset,
					      SECTOR_SIZE);
	}
	if (block_size % UDS_BLOCK_SIZE != 0) {
		return uds_log_error_strerror(
			UDS_INCORRECT_ALIGNMENT,
			"block_size %zd not multiple of %d",
			block_size,
			UDS_BLOCK_SIZE);
	}

#ifdef DM_BUFIO_CLIENT_NO_SLEEP
	client = dm_bufio_client_create(
		factory->bdev, block_size, reserved_buffers, 0, NULL, NULL, 0);
#else
	client = dm_bufio_client_create(
		factory->bdev, block_size, reserved_buffers, 0, NULL, NULL);
#endif
	if (IS_ERR(client)) {
		return -PTR_ERR(client);
	}

	dm_bufio_set_sector_offset(client, offset >> SECTOR_SHIFT);
	*client_ptr = client;
	return UDS_SUCCESS;
}

int open_uds_buffered_reader(struct io_factory *factory,
			     off_t offset,
			     size_t size,
			     struct buffered_reader **reader_ptr)
{
	int result;
	struct dm_bufio_client *client = NULL;

	if (size % UDS_BLOCK_SIZE != 0) {
		return uds_log_error_strerror(
			UDS_INCORRECT_ALIGNMENT,
			"region size %zd is not multiple of %d",
			size,
			UDS_BLOCK_SIZE);
	}

	result = make_uds_bufio(factory, offset, UDS_BLOCK_SIZE, 1, &client);
	if (result != UDS_SUCCESS) {
		return result;
	}

	result = make_buffered_reader(
		factory, client, size / UDS_BLOCK_SIZE, reader_ptr);
	if (result != UDS_SUCCESS) {
		dm_bufio_client_destroy(client);
	}
	return result;
}

int open_uds_buffered_writer(struct io_factory *factory,
			     off_t offset,
			     size_t size,
			     struct buffered_writer **writer_ptr)
{
	int result;
	struct dm_bufio_client *client = NULL;

	if (size % UDS_BLOCK_SIZE != 0) {
		return uds_log_error_strerror(UDS_INCORRECT_ALIGNMENT,
					      "region size %zd is not multiple of %d",
					      size,
					      UDS_BLOCK_SIZE);
	}

	result = make_uds_bufio(factory, offset, UDS_BLOCK_SIZE, 1, &client);
	if (result != UDS_SUCCESS) {
		return result;
	}

	result = make_buffered_writer(
		factory, client, size / UDS_BLOCK_SIZE, writer_ptr);
	if (result != UDS_SUCCESS) {
		dm_bufio_client_destroy(client);
	}
	return result;
}
